from typing import TypeVar, Generic, List, NamedTuple, Callable

from six.moves import range

from trashcli.lib.my_input import Input
from trashcli.restore.index import Sequence
from trashcli.restore.output import Output
from trashcli.restore.output_event import Die, OutputEvent, Quit
from trashcli.restore.output_event import Exiting
from trashcli.restore.range import Range
from trashcli.restore.sequences import Sequences
from trashcli.restore.single import Single
from trashcli.restore.trashed_file import TrashedFile

SelectedFiles = NamedTuple('SelectedFiles', [
    ('files_to_restore', List[TrashedFile]),
    ('overwrite', bool),
])

Context = NamedTuple('Context', [
    ('trashed_files', List[TrashedFile]),
    ('overwrite', bool),
])

InputRead = NamedTuple('InputRead', [
    ('user_input', str),
    ('trashed_files', List[TrashedFile]),
    ('overwrite', bool),
])


class RestoreAskingTheUser(object):
    def __init__(self,
                 input,  # type: Input
                 restorer,
                 output,  # type: Output
                 ):
        self.input = input
        self.restorer = restorer
        self.output = output

    def read_user_input(self,
                        args,  # type: Context
                        ):  # type: (...) -> Either[OutputEvent, InputRead]
        try:
            user_input = self.input.read_input(
                "What file to restore [0..%d]: " % (
                        len(args.trashed_files) - 1))

        except KeyboardInterrupt:
            return Left(Quit())
        except EOFError:
            return Left(Quit())
        else:
            if user_input == "":
                return Left(Exiting("No files were restored"))
            else:
                return Right(
                    InputRead(user_input, args.trashed_files, args.overwrite))

    def restore_asking_the_user(self, trashed_files, overwrite):
        input = Right(Context(trashed_files, overwrite))
        compose(input, [
            self.read_user_input,
            trashed_files_to_restore,
            self.restore_selected_files,
        ]).on_error(
            lambda error: self.output.append_event(error))

    def restore_selected_files(self,
                               selected_files,  # type: SelectedFiles
                               ):  # type: (...) -> Either[Die, None]
        try:
            for trashed_file in selected_files.files_to_restore:
                self.restorer.restore_trashed_file(trashed_file,
                                                   selected_files.overwrite)
            return Right(None)
        except IOError as e:
            return Left(Die(e))


Error = TypeVar('Error')
Value = TypeVar('Value')


def compose(input, funcs):
    for f in funcs:
        input = input.apply(f)
    return input


class Either(Generic[Error, Value]):
    def apply(self,
              f):  # type: (Callable[[Value], Either[Error, OutputValue]]) -> Either[Error, OutputValue]
        raise NotImplementedError()


class Left(Either, Generic[Error, Value]):
    def __init__(self,
                 error,  # type: Error
                 ):
        self.error = error

    def apply(self,
              f):  # type: (Callable[[Value], Either[Error, OutputValue]]) -> Either[Error, OutputValue]
        return self

    def on_error(self, f):
        return f(self.error)


OutputValue = TypeVar('OutputValue')


class Right(Either[Error, Value]):
    def __init__(self, value,  # type: Value
                 ):
        self.value = value

    def apply(self,
              f):  # type: (Callable[[Value], Either[Error, OutputValue]]) -> Either[Error, OutputValue]
        return f(self.value)

    def on_error(self, f):
        return self


def trashed_files_to_restore(input_read,  # type: InputRead
                             ):  # type: (...) -> Either[Die, SelectedFiles]
    try:
        sequences = parse_indexes(input_read.user_input,
                                  len(input_read.trashed_files))
        file_to_restore = [input_read.trashed_files[index] for index in
                           sequences.all_indexes()]
        selected_files = SelectedFiles(file_to_restore, input_read.overwrite)
        return Right(selected_files)
    except InvalidEntry as e:
        return Left(Die("Invalid entry: %s" % e))


class InvalidEntry(Exception):
    pass


def parse_indexes(user_input,  # type: str
                  len_trashed_files,  # type: int
                  ):  # type: (...) -> Sequences
    indexes = user_input.split(',')
    sequences = []  # type: List[Sequence]
    for index in indexes:
        if "-" in index:
            first, last = index.split("-", 2)
            if first == "" or last == "":
                raise InvalidEntry("open interval: %s" % index)
            split = list(map(parse_int_index, (first, last)))
            sequences.append(Range(split[0], split[1]))
        else:
            int_index = parse_int_index(index)
            sequences.append(Single(int_index))
    result = Sequences(sequences)
    acceptable_values = range(0, len_trashed_files)
    for index in result.all_indexes():
        if not index in acceptable_values:
            raise InvalidEntry(
                "out of range %s..%s: %s" %
                (acceptable_values[0], acceptable_values[-1], index))
    return result


def parse_int_index(text):  # type: (str) -> int
    try:
        return int(text)
    except ValueError:
        raise InvalidEntry("not an index: %s" % text)



from typing import Iterable
from typing import NamedTuple
from typing import Optional
from typing import Union

from trashcli.lib.path_of_backup_copy import path_of_backup_copy
from trashcli.parse_trashinfo.parse_deletion_date import parse_deletion_date
from trashcli.parse_trashinfo.parse_original_location import \
    parse_original_location
from trashcli.restore.file_system import FileReader
from trashcli.restore.info_dir_searcher import InfoDirSearcher
from trashcli.restore.restore_logger import RestoreLogger
from trashcli.restore.trashed_file import TrashedFile


class TrashedFiles:
    def __init__(self,
                 logger,  # type: RestoreLogger
                 file_reader,  # type: FileReader
                 searcher,  # type: InfoDirSearcher
                 ):
        self.logger = logger
        self.file_reader = file_reader
        self.searcher = searcher

    def all_trashed_files(self,
                          trash_dir_from_cli,  # type: Optional[str]
                          ):  # type: (...) -> Iterable[TrashedFile]
        for event in self.all_trashed_files_internal(trash_dir_from_cli):
            if type(event) is NonTrashinfoFileFound:
                self.logger.warning("Non .trashinfo file in info dir")
            elif type(event) is NonParsableTrashInfo:
                self.logger.warning(
                    "Non parsable trashinfo file: %s, because %s" %
                    (event.path, event.reason))
            elif type(event) is IOErrorReadingTrashInfo:
                self.logger.warning(str(event))
            elif type(event) is TrashedFileFound:
                yield event.trashed_file
            else:
                raise RuntimeError()

    def all_trashed_files_internal(self,
                                   trash_dir_from_cli,  # type: Optional[str]
                                   ):  # type: (...) -> Iterable[Event]
        for info_file in self.searcher.all_file_in_info_dir(trash_dir_from_cli):
            if info_file.type == 'non_trashinfo':
                yield NonTrashinfoFileFound(info_file.path)
            elif info_file.type == 'trashinfo':
                try:
                    contents = self.file_reader.contents_of(info_file.path)
                    original_location = parse_original_location(contents,
                                                                info_file.volume)
                    deletion_date = parse_deletion_date(contents)
                    backup_file_path = path_of_backup_copy(info_file.path)
                    trashedfile = TrashedFile(original_location,
                                              deletion_date,
                                              info_file.path,
                                              backup_file_path)
                    yield TrashedFileFound(trashedfile)
                except ValueError as e:
                    yield NonParsableTrashInfo(info_file.path, e)
                except IOError as e:
                    yield IOErrorReadingTrashInfo(info_file.path, str(e))
            else:
                raise RuntimeError("Unexpected file type: %s: %s",
                                   info_file.type, info_file.path)


class NonTrashinfoFileFound(
    NamedTuple('NonTrashinfoFileFound', [
        ('path', str),
    ])): pass


class TrashedFileFound(
    NamedTuple('TrashedFileFound', [
        ('trashed_file', TrashedFile),
    ])): pass


class NonParsableTrashInfo(
    NamedTuple('NonParsableTrashInfo', [
        ('path', str),
        ('reason', Exception),
    ])): pass


class IOErrorReadingTrashInfo(
    NamedTuple('IOErrorReadingTrashInfo', [
        ('path', str),
        ('error', str),
    ])): pass


Event = Union[
    NonTrashinfoFileFound,
    TrashedFileFound,
    NonParsableTrashInfo,
    IOErrorReadingTrashInfo]



from pprint import pprint


class DebugVolumesArgs:
    pass


class DebugVolumes(object):
    def run_action(self,
                   _args,  # type: DebugVolumesArgs
                   ):
        import psutil
        import os
        all = sorted([p for p in psutil.disk_partitions(all=True)],
                     key=lambda p: p.device)
        physical = sorted([p for p in psutil.disk_partitions()],
                          key=lambda p: p.device)
        virtual = [p for p in all if p not in physical]
        print("physical ->")
        pprint(physical)
        print("virtual ->")
        pprint(virtual)
        os.system('df -P')



import os

from trashcli.fstab.volume_listing import RealIsMount
from trashcli.fstab.volume_of import VolumeOf
from trashcli.fstab.volume_of_impl import VolumeOfImpl


class RealVolumeOf(VolumeOf):
    def __init__(self):
        self.impl = VolumeOfImpl(RealIsMount(), os.path.abspath)

    def volume_of(self, path):
        return self.impl.volume_of(path)



from abc import ABCMeta, abstractmethod

import six

from trashcli.restore.output_event import OutputEvent


@six.add_metaclass(ABCMeta)
class Output:
    @abstractmethod
    def quit(self):
        raise NotImplementedError

    @abstractmethod
    def die(self, msg):
        raise NotImplementedError

    @abstractmethod
    def println(self, msg):
        raise NotImplementedError

    @abstractmethod
    def append_event(self,
                     event,  # type: OutputEvent
                     ):
        raise NotImplementedError



from __future__ import absolute_import

from six.moves.urllib.parse import unquote

from trashcli.parse_trashinfo.parser_error import ParseError


def parse_path(contents):
    for line in contents.split('\n'):
        if line.startswith('Path='):
            return unquote(line[len('Path='):])
    raise ParseError('Unable to parse Path')
